class FIXError(Exception):
    pass


class FIXMessageError(FIXError):
    pass


class EncodingError(FIXError):
    pass


class DecodingError(FIXError):
    pass


class TagNotFoundError(FIXMessageError):
    pass


class DuplicatedTagError(FIXMessageError):
    pass


class RepeatingTagError(FIXMessageError):
    pass


class UnmappedRepeatedGrpError(FIXMessageError):
    pass



import asyncio
import unittest
from unittest.mock import AsyncMock, MagicMock, patch

from asyncfix import FMsg, FTag
from asyncfix.codec import Codec
from asyncfix.connection import FIXConnectionHandler
from asyncfix.engine import FIXEngine
from asyncfix.message import FIXContainer, FIXMessage
from asyncfix.protocol import FIXProtocol44
from asyncfix.session import FIXSession


class TestConnection(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.engine = FIXEngine()
        self.protocol = FIXProtocol44()
        self.codec = Codec(self.protocol)
        self.session = FIXSession("test", "test_target", "test_sender")

        self.socket_reader_mock = MagicMock()
        self.socket_reader_mock.read = AsyncMock()
        msg = FIXMessage(FMsg.NEWORDERSINGLE)
        msg.set(FTag.Price, "123.45")
        msg.set(FTag.OrderQty, 9876)
        msg.set(FTag.Symbol, "VOD.L")
        self.msg = msg

    async def test_handle_read(self):
        self.socket_reader_mock.read.side_effect = [
            self.codec.encode(self.msg, self.session).encode(),
            self.codec.encode(self.msg, self.session).encode(),
            asyncio.CancelledError,  # This stops infinite loop
        ]

        h = FIXConnectionHandler(
            self.engine, self.protocol, self.socket_reader_mock, None
        )

        with patch.object(h, "process_message") as mock_process_message:
            await asyncio.wait_for(h.handle_read(), 1)

        self.assertEqual(mock_process_message.call_count, 2)

        for args in mock_process_message.call_args_list:
            self.assertEqual(len(args[0]), 1)  # 1 arg
            self.assertEqual(args[1], {})  # no kwargs

            assert {
                FTag.Symbol: "VOD.L",
                FTag.Price: "123.45",
                FTag.OrderQty: "9876",
            } in args[0][0]

    async def test_handle_read_partial_msg(self):
        full = self.codec.encode(self.msg, self.session).encode()
        partial_1 = full[:20]
        partial_2 = full[20:]
        self.assertEqual((partial_1 + partial_2), full)

        self.socket_reader_mock.read.side_effect = [
            self.codec.encode(self.msg, self.session).encode(),
            partial_1,
            partial_2,
            asyncio.CancelledError,  # This stops infinite loop
        ]

        h = FIXConnectionHandler(
            self.engine, self.protocol, self.socket_reader_mock, None
        )

        with patch.object(h, "process_message") as mock_process_message:
            await asyncio.wait_for(h.handle_read(), 1)

        self.assertEqual(mock_process_message.call_count, 2)

        for args in mock_process_message.call_args_list:
            self.assertEqual(len(args[0]), 1)  # 1 arg
            self.assertEqual(args[1], {})  # no kwargs

            assert {
                FTag.Symbol: "VOD.L",
                FTag.Price: "123.45",
                FTag.OrderQty: "9876",
            } in args[0][0]

    async def test_handle_read_partial_bad_data(self):
        full = self.codec.encode(self.msg, self.session).encode()
        partial_1 = full[:20]
        partial_2 = full[20:]
        self.assertEqual((partial_1 + partial_2), full)

        self.socket_reader_mock.read.side_effect = [
            self.codec.encode(self.msg, self.session).encode(),
            partial_1,
            partial_1,
            asyncio.CancelledError,  # This stops infinite loop
        ]

        h = FIXConnectionHandler(
            self.engine, self.protocol, self.socket_reader_mock, None
        )

        with patch.object(h, "process_message") as mock_process_message:
            await asyncio.wait_for(h.handle_read(), 1)

        self.assertEqual(mock_process_message.call_count, 1)

        for args in mock_process_message.call_args_list:
            self.assertEqual(len(args[0]), 1)  # 1 arg
            self.assertEqual(args[1], {})  # no kwargs

            assert {
                FTag.Symbol: "VOD.L",
                FTag.Price: "123.45",
                FTag.OrderQty: "9876",
            } in args[0][0]

    async def test_handle_read_partial_decode_logic_msg_buff_management(self):
        full = self.codec.encode(self.msg, self.session).encode()
        partial_1 = full[:20]
        partial_2 = full[20:]
        self.assertEqual((partial_1 + partial_2), full)

        msg_valid = b"8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa

        h = FIXConnectionHandler(
            self.engine, self.protocol, self.socket_reader_mock, None
        )
        with (patch.object(h, "process_message") as mock_process_message,):
            self.socket_reader_mock.read.side_effect = [
                b"abcd",
                msg_valid,
                asyncio.CancelledError,  # This stops infinite loop
            ]
            await asyncio.wait_for(h.handle_read(), 1)

        self.assertEqual(mock_process_message.call_count, 1)

    async def test_handle_read_partial_msg_oneafter_another(self):
        full = self.codec.encode(self.msg, self.session).encode()
        partial_1 = full[:20]
        partial_2 = full[20:]
        self.assertEqual((partial_1 + partial_2), full)

        self.socket_reader_mock.read.side_effect = [
            partial_1,
            # remainder comes with next valid message
            partial_2 + self.codec.encode(self.msg, self.session).encode(),
            asyncio.CancelledError,  # This stops infinite loop
        ]

        h = FIXConnectionHandler(
            self.engine, self.protocol, self.socket_reader_mock, None
        )

        with patch.object(h, "process_message") as mock_process_message:
            await asyncio.wait_for(h.handle_read(), 1)

        self.assertEqual(mock_process_message.call_count, 2)

        for args in mock_process_message.call_args_list:
            self.assertEqual(len(args[0]), 1)  # 1 arg
            self.assertEqual(args[1], {})  # no kwargs

            assert {
                FTag.Symbol: "VOD.L",
                FTag.Price: "123.45",
                FTag.OrderQty: "9876",
            } in args[0][0]



import asyncio
import logging

from asyncfix import FMsg, FTag
from asyncfix.connection import ConnectionState, FIXConnectionHandler, FIXEndPoint
from asyncfix.engine import FIXEngine
from asyncfix.journaler import DuplicateSeqNoError
from asyncfix.message import FIXMessage
from asyncfix.protocol import FIXProtocolBase


class FIXClientConnectionHandler(FIXConnectionHandler):
    def __init__(
        self,
        engine: FIXEngine,
        protocol: FIXProtocolBase,
        target_comp_id: str,
        sender_comp_id: str,
        socket_reader: asyncio.StreamReader,
        socket_writer: asyncio.StreamWriter,
        addr=None,
        observer=None,
        target_sub_id=None,
        sender_sub_id=None,
        heartbeat_timeout=30,
        heartbeat=1,
    ):
        super().__init__(
            engine=engine,
            protocol=protocol,
            socket_reader=socket_reader,
            socket_writer=socket_writer,
            addr=addr,
            observer=observer,
        )

        self.target_comp_id = target_comp_id
        self.sender_comp_id = sender_comp_id
        self.target_sub_id = target_sub_id
        self.sender_sub_id = sender_sub_id
        self.heartbeat_period = float(heartbeat_timeout)
        self.heartbeat = heartbeat

        # we need to send a login request.
        self.session = self.engine.get_or_create_session_from_comp_ids(
            self.target_comp_id, self.sender_comp_id
        )
        if self.session is None:
            raise RuntimeError("Failed to create client session")

        self.protocol = protocol

        asyncio.ensure_future(self.logon())

    async def logon(self):
        logon_msg = self.protocol.logon()
        logon_msg.set(FTag.HeartBtInt, self.heartbeat, replace=True)
        await self.send_msg(logon_msg)

    async def handle_session_message(self, msg: FIXMessage):
        responses = []

        recv_seq_no = msg[FTag.MsgSeqNum]

        msg_type = msg[FTag.MsgType]
        target_comp_d = msg[FTag.TargetCompID]
        sender_comp_id = msg[FTag.SenderCompID]

        if msg_type == FMsg.LOGON:
            if self.connection_state == ConnectionState.LOGGED_IN:
                logging.warning(
                    "Client session already logged in - ignoring login request"
                )
            else:
                try:
                    self.connection_state = ConnectionState.LOGGED_IN
                    self.heartbeat_period = float(msg[FTag.HeartBtInt])
                except DuplicateSeqNoError:
                    logging.error(
                        "Failed to process login request with duplicate seq no"
                    )
                    await self.disconnect()
                    return
        elif self.connection_state == ConnectionState.LOGGED_IN:
            # compids are reversed here
            if not self.session.validate_comp_ids(sender_comp_id, target_comp_d):
                logging.error("Received message with unexpected comp ids")
                await self.disconnect()
                return

            if msg_type == FMsg.LOGOUT:
                self.connection_state = ConnectionState.LOGGED_OUT
                self.handle_close()
            elif msg_type == FMsg.TESTREQUEST:
                responses.append(self.protocol.heartbeat())
            elif msg_type == FMsg.RESENDREQUEST:
                responses.extend(self._handle_resend_request(msg))
            elif msg_type == FMsg.SEQUENCERESET:
                # we can treat GapFill and SequenceReset in the same way
                # in both cases we will just reset the seq number to the
                # NewSeqNo received in the message
                new_seq_no = msg[FTag.NewSeqNo]
                if msg[FTag.GapFillFlag] == "Y":
                    logging.info(
                        "Received SequenceReset(GapFill) filling gap from %s to %s"
                        % (recv_seq_no, new_seq_no)
                    )
                self.session.set_recv_seq_no(int(new_seq_no) - 1)
                recv_seq_no = new_seq_no
        else:
            logging.warning("Can't process message, counterparty is not logged in")

        return (recv_seq_no, responses)


class FIXClient(FIXEndPoint):
    def __init__(
        self,
        engine: FIXEngine,
        protocol: FIXProtocolBase,
        target_comp_id,
        sender_comp_id,
        target_sub_id=None,
        sender_sub_id=None,
        heartbeat_timeout=30,
        with_seq_no_reset=True,
    ):
        self.target_comp_id = target_comp_id
        self.sender_comp_id = sender_comp_id
        self.target_sub_id = target_sub_id
        self.sender_sub_id = sender_sub_id
        self.heartbeat_timeout = heartbeat_timeout
        self.with_seq_no_reset = with_seq_no_reset
        self.socket_reader = self.socket_writer = None
        self.addr = None

        FIXEndPoint.__init__(self, engine, protocol)

    async def start(self, host, port):
        self.socket_reader, self.socket_writer = await asyncio.open_connection(
            host, port
        )
        self.addr = (host, port)

        logging.info("Connected to %s" % repr(self.addr))
        connection = FIXClientConnectionHandler(
            engine=self.engine,
            protocol=self.protocol,
            target_comp_id=self.target_comp_id,
            sender_comp_id=self.sender_comp_id,
            socket_reader=self.socket_reader,
            socket_writer=self.socket_writer,
            addr=self.addr,
            observer=self,
            target_sub_id=self.target_sub_id,
            sender_sub_id=self.sender_sub_id,
            heartbeat_timeout=self.heartbeat_timeout,
        )
        asyncio.create_task(connection.handle_read())

        self.connections.append(connection)

        for handler in filter(
            lambda x: x[1] == ConnectionState.CONNECTED, self.message_handlers
        ):
            await handler[0](connection)

    def stop(self):
        logging.info("Stopping client connections")
        for connection in self.connections:
            connection.disconnect()
        self.socket_writer.close()



from math import isnan, nan

from asyncfix import FIXMessage, FMsg, FTag

from .common import FExecType, FOrdStatus
from .order_single import FIXNewOrderSingle
from .schema import FIXSchema


class FIXTester:
    def __init__(self, schema: FIXSchema | None = None):
        self.registered_orders = {}
        self.schema = schema
        self.order_id = 0
        self.exec_id = 10000

    def next_order_id(self) -> int:
        self.order_id += 1
        return self.order_id

    def next_exec_id(self) -> int:
        self.exec_id += 1
        return self.exec_id

    def order_register_single(self, o: FIXNewOrderSingle):
        self.registered_orders[o.clord_id] = o
        return True

    def fix_cxl_request(self, o: FIXNewOrderSingle) -> FIXMessage:
        m = o.cancel_req()
        if self.schema:
            self.schema.validate(m)
        assert o.can_cancel()  # new assert 2023-09-23
        o.status = FOrdStatus.PENDING_CANCEL
        self.registered_orders[o.clord_id] = o
        return m

    def fix_rep_request(
        self, o: FIXNewOrderSingle, price: float = nan, qty: float = nan
    ) -> FIXMessage:
        m = o.replace_req(price, qty)
        if self.schema:
            self.schema.validate(m)
        assert o.can_replace()
        o.status = FOrdStatus.PENDING_REPLACE
        self.registered_orders[o.clord_id] = o
        return m

    def fix_cxlrep_reject_msg(
        self,
        cxl_req: FIXMessage,
        ord_status: FOrdStatus,
    ) -> FIXMessage:
        clord_id = cxl_req[FTag.ClOrdID]
        orig_clord_id = cxl_req[FTag.OrigClOrdID]

        m = FIXMessage(FMsg.ORDERCANCELREJECT)
        m[37] = 0
        m[11] = clord_id
        m[41] = orig_clord_id
        m[39] = ord_status

        assert cxl_req.msg_type in [
            FMsg.ORDERCANCELREQUEST,
            FMsg.ORDERCANCELREPLACEREQUEST,
        ]

        if cxl_req.msg_type == FMsg.ORDERCANCELREQUEST:
            m[FTag.CxlRejResponseTo] = "1"
        elif cxl_req.msg_type == FMsg.ORDERCANCELREPLACEREQUEST:
            m[FTag.CxlRejResponseTo] = "2"

        if self.schema:
            self.schema.validate(m)

        return m

    def fix_exec_report_msg(
        self,
        order: FIXNewOrderSingle,
        clord_id: str,
        exec_type: FExecType,
        ord_status: FOrdStatus,
        cum_qty: float = nan,
        leaves_qty: float = nan,
        last_qty: float = nan,
        price: float = nan,
        order_qty: float = nan,
        orig_clord_id: str = None,
    ) -> FIXMessage:
        assert order.clord_id in self.registered_orders, "Unregistered order!"

        m = FIXMessage(FMsg.EXECUTIONREPORT)
        assert clord_id
        m[FTag.ClOrdID] = clord_id

        if order.order_id is None:
            order_id = self.next_order_id()
        else:
            order_id = order.order_id

        m[FTag.OrderID] = order_id
        m[FTag.ExecID] = self.next_exec_id()

        if orig_clord_id:
            m[FTag.OrigClOrdID] = orig_clord_id
        m[FTag.ExecType] = exec_type
        m[FTag.OrdStatus] = ord_status
        m[FTag.Side] = order.side

        if isnan(order_qty):
            order_qty = order.qty
        else:
            assert (
                exec_type == FExecType.REPLACED
            ), "Only applicable to exec_type=5 (replace)"
            assert order_qty > 0

        if isnan(cum_qty):
            cum_qty = order.cum_qty
        else:
            assert cum_qty <= order.qty
            assert cum_qty >= 0
        m[FTag.CumQty] = cum_qty

        if isnan(leaves_qty):
            leaves_qty = order.leaves_qty
        else:
            assert leaves_qty >= 0
            assert leaves_qty <= order_qty

        m[FTag.LeavesQty] = leaves_qty
        assert (
            cum_qty + leaves_qty <= order_qty
        ), f"cum_qty[{cum_qty}] + leaves_qty[{leaves_qty}] <= order_qty[{order_qty}]"

        if not isnan(last_qty):
            assert not isnan(leaves_qty), "Must also set leaves_qty, when trade"
            assert not isnan(cum_qty), "Must also set cum_qty, when trade"
            assert (
                exec_type == FExecType.TRADE
            ), "Only applicable to exec_type=F (trade)"
            assert last_qty > 0
            m[FTag.LastQty] = last_qty
            assert (
                round(last_qty - (cum_qty - order.cum_qty), 3) == 0
            ), "Probably incorrect Trade qty"
        else:
            assert (
                exec_type != FExecType.TRADE
            ), "You must set last_qty when exec_type=F (trade)"

        if not isnan(price):
            assert (
                exec_type == FExecType.REPLACED
            ), "Only applicable to exec_type=5 (replace)"
        else:
            price = order.price

        order.set_instrument(m)

        order.set_price_qty(m, price, order_qty)
        m[FTag.AvgPx] = price

        order.set_account(m)

        if (
            exec_type == FExecType.PENDING_CANCEL
            and ord_status == FOrdStatus.PENDING_CANCEL
        ):
            assert order.orig_clord_id
            assert order.clord_id
            assert order.clord_id != order.orig_clord_id
            assert order.cum_qty == cum_qty
            assert order.leaves_qty == leaves_qty

        if (
            ord_status == FOrdStatus.FILLED
            or ord_status == FOrdStatus.CANCELED
            or ord_status == FOrdStatus.REJECTED
            or ord_status == FOrdStatus.EXPIRED
        ):
            assert leaves_qty == 0, "New order report is finished, but LeavesQty != 0"

        if self.schema:
            self.schema.validate(m)
        return m



from asyncfix.journaler import Journaler
from asyncfix.message import FIXMessage
from asyncfix.session import FIXSession


class FIXEngine(object):
    def __init__(self, journalfile=None):
        self.journaller = Journaler(journalfile)
        self.sessions: dict[str, FIXSession] = {}

        # We load all sessions from the journal and add to our list
        for session in self.journaller.sessions():
            self.sessions[session.key] = session

    def validate_session(self, target_comp_id, sender_comp_id):
        # this make any session we receive valid
        return True

    def should_resend_message(self, session: FIXSession, msg: FIXMessage):
        # we should resend all application messages
        return True

    def create_session(self, target_comp_id, sender_comp_id):
        if self.find_session_by_comp_ids(target_comp_id, sender_comp_id) is None:
            session = self.journaller.create_session(target_comp_id, sender_comp_id)
            self.sessions[session.key] = session
        else:
            raise RuntimeError("Failed to add session with duplicate key")
        return session

    def get_session(self, identifier):
        try:
            return self.sessions[identifier]
        except KeyError:
            return None

    def find_session_by_comp_ids(self, target_comp_id, sender_comp_id):
        sessions = [
            x
            for x in self.sessions.values()
            if x.target_comp_id == target_comp_id and x.sender_comp_id == sender_comp_id
        ]
        if sessions is not None and len(sessions) != 0:
            return sessions[0]
        return None

    def get_or_create_session_from_comp_ids(self, target_comp_id, sender_comp_id):
        session = self.find_session_by_comp_ids(target_comp_id, sender_comp_id)
        if session is None:
            if self.validate_session(target_comp_id, sender_comp_id):
                session = self.create_session(target_comp_id, sender_comp_id)

        return session



import enum


class TagEnumMeta(enum.EnumMeta):
    def __contains__(cls, item):
        if isinstance(item, enum.Enum):
            return isinstance(item, cls) and item._name_ in cls._member_map_
        else:
            v = str(item)
            return v in cls._value2member_map_


class FTag(enum.Enum, metaclass=TagEnumMeta):
    def __str__(self):
        return str(self.value)

    def __eq__(self, o):
        return self.value == str(o)

    def __hash__(self):
        return hash(self.value)

    Account = "1"
    AdvId = "2"
    AdvRefID = "3"
    AdvSide = "4"
    AdvTransType = "5"
    AvgPx = "6"
    BeginSeqNo = "7"
    BeginString = "8"
    BodyLength = "9"
    CheckSum = "10"
    ClOrdID = "11"
    Commission = "12"
    CommType = "13"
    CumQty = "14"
    Currency = "15"
    EndSeqNo = "16"
    ExecID = "17"
    ExecInst = "18"
    ExecRefID = "19"
    ExecTransType = "20"
    HandlInst = "21"
    SecurityIDSource = "22"
    IOIID = "23"
    IOIOthSvc = "24"
    IOIQltyInd = "25"
    IOIRefID = "26"
    IOIQty = "27"
    IOITransType = "28"
    LastCapacity = "29"
    LastMkt = "30"
    LastPx = "31"
    LastQty = "32"
    NoLinesOfText = "33"
    MsgSeqNum = "34"
    MsgType = "35"
    NewSeqNo = "36"
    OrderID = "37"
    OrderQty = "38"
    OrdStatus = "39"
    OrdType = "40"
    OrigClOrdID = "41"
    OrigTime = "42"
    PossDupFlag = "43"
    Price = "44"
    RefSeqNum = "45"
    RelatdSym = "46"
    Rule80A = "47"
    SecurityID = "48"
    SenderCompID = "49"
    SenderSubID = "50"
    SendingDate = "51"
    SendingTime = "52"
    Quantity = "53"
    Side = "54"
    Symbol = "55"
    TargetCompID = "56"
    TargetSubID = "57"
    Text = "58"
    TimeInForce = "59"
    TransactTime = "60"
    Urgency = "61"
    ValidUntilTime = "62"
    SettlType = "63"
    SettlDate = "64"
    SymbolSfx = "65"
    ListID = "66"
    ListSeqNo = "67"
    TotNoOrders = "68"
    ListExecInst = "69"
    AllocID = "70"
    AllocTransType = "71"
    RefAllocID = "72"
    NoOrders = "73"
    AvgPxPrecision = "74"
    TradeDate = "75"
    ExecBroker = "76"
    PositionEffect = "77"
    NoAllocs = "78"
    AllocAccount = "79"
    AllocQty = "80"
    ProcessCode = "81"
    NoRpts = "82"
    RptSeq = "83"
    CxlQty = "84"
    NoDlvyInst = "85"
    DlvyInst = "86"
    AllocStatus = "87"
    AllocRejCode = "88"
    Signature = "89"
    SecureDataLen = "90"
    SecureData = "91"
    BrokerOfCredit = "92"
    SignatureLength = "93"
    EmailType = "94"
    RawDataLength = "95"
    RawData = "96"
    PossResend = "97"
    EncryptMethod = "98"
    StopPx = "99"
    ExDestination = "100"
    CxlRejReason = "102"
    OrdRejReason = "103"
    IOIQualifier = "104"
    WaveNo = "105"
    Issuer = "106"
    SecurityDesc = "107"
    HeartBtInt = "108"
    ClientID = "109"
    MinQty = "110"
    MaxFloor = "111"
    TestReqID = "112"
    ReportToExch = "113"
    LocateReqd = "114"
    OnBehalfOfCompID = "115"
    OnBehalfOfSubID = "116"
    QuoteID = "117"
    NetMoney = "118"
    SettlCurrAmt = "119"
    SettlCurrency = "120"
    ForexReq = "121"
    OrigSendingTime = "122"
    GapFillFlag = "123"
    NoExecs = "124"
    CxlType = "125"
    ExpireTime = "126"
    DKReason = "127"
    DeliverToCompID = "128"
    DeliverToSubID = "129"
    IOINaturalFlag = "130"
    QuoteReqID = "131"
    BidPx = "132"
    OfferPx = "133"
    BidSize = "134"
    OfferSize = "135"
    NoMiscFees = "136"
    MiscFeeAmt = "137"
    MiscFeeCurr = "138"
    MiscFeeType = "139"
    PrevClosePx = "140"
    ResetSeqNumFlag = "141"
    SenderLocationID = "142"
    TargetLocationID = "143"
    OnBehalfOfLocationID = "144"
    DeliverToLocationID = "145"
    NoRelatedSym = "146"
    Subject = "147"
    Headline = "148"
    URLLink = "149"
    ExecType = "150"
    LeavesQty = "151"
    CashOrderQty = "152"
    AllocAvgPx = "153"
    AllocNetMoney = "154"
    SettlCurrFxRate = "155"
    SettlCurrFxRateCalc = "156"
    NumDaysInterest = "157"
    AccruedInterestRate = "158"
    AccruedInterestAmt = "159"
    SettlInstMode = "160"
    AllocText = "161"
    SettlInstID = "162"
    SettlInstTransType = "163"
    EmailThreadID = "164"
    SettlInstSource = "165"
    SettlLocation = "166"
    SecurityType = "167"
    EffectiveTime = "168"
    StandInstDbType = "169"
    StandInstDbName = "170"
    StandInstDbID = "171"
    SettlDeliveryType = "172"
    SettlDepositoryCode = "173"
    SettlBrkrCode = "174"
    SettlInstCode = "175"
    SecuritySettlAgentName = "176"
    SecuritySettlAgentCode = "177"
    SecuritySettlAgentAcctNum = "178"
    SecuritySettlAgentAcctName = "179"
    SecuritySettlAgentContactName = "180"
    SecuritySettlAgentContactPhone = "181"
    CashSettlAgentName = "182"
    CashSettlAgentCode = "183"
    CashSettlAgentAcctNum = "184"
    CashSettlAgentAcctName = "185"
    CashSettlAgentContactName = "186"
    CashSettlAgentContactPhone = "187"
    BidSpotRate = "188"
    BidForwardPoints = "189"
    OfferSpotRate = "190"
    OfferForwardPoints = "191"
    OrderQty2 = "192"
    SettlDate2 = "193"
    LastSpotRate = "194"
    LastForwardPoints = "195"
    AllocLinkID = "196"
    AllocLinkType = "197"
    SecondaryOrderID = "198"
    NoIOIQualifiers = "199"
    MaturityMonthYear = "200"
    PutOrCall = "201"
    StrikePrice = "202"
    CoveredOrUncovered = "203"
    CustomerOrFirm = "204"
    MaturityDay = "205"
    OptAttribute = "206"
    SecurityExchange = "207"
    NotifyBrokerOfCredit = "208"
    AllocHandlInst = "209"
    MaxShow = "210"
    PegOffsetValue = "211"
    XmlDataLen = "212"
    XmlData = "213"
    SettlInstRefID = "214"
    NoRoutingIDs = "215"
    RoutingType = "216"
    RoutingID = "217"
    Spread = "218"
    Benchmark = "219"
    BenchmarkCurveCurrency = "220"
    BenchmarkCurveName = "221"
    BenchmarkCurvePoint = "222"
    CouponRate = "223"
    CouponPaymentDate = "224"
    IssueDate = "225"
    RepurchaseTerm = "226"
    RepurchaseRate = "227"
    Factor = "228"
    TradeOriginationDate = "229"
    ExDate = "230"
    ContractMultiplier = "231"
    NoStipulations = "232"
    StipulationType = "233"
    StipulationValue = "234"
    YieldType = "235"
    Yield = "236"
    TotalTakedown = "237"
    Concession = "238"
    RepoCollateralSecurityType = "239"
    RedemptionDate = "240"
    UnderlyingCouponPaymentDate = "241"
    UnderlyingIssueDate = "242"
    UnderlyingRepoCollateralSecurityType = "243"
    UnderlyingRepurchaseTerm = "244"
    UnderlyingRepurchaseRate = "245"
    UnderlyingFactor = "246"
    UnderlyingRedemptionDate = "247"
    LegCouponPaymentDate = "248"
    LegIssueDate = "249"
    LegRepoCollateralSecurityType = "250"
    LegRepurchaseTerm = "251"
    LegRepurchaseRate = "252"
    LegFactor = "253"
    LegRedemptionDate = "254"
    CreditRating = "255"
    UnderlyingCreditRating = "256"
    LegCreditRating = "257"
    TradedFlatSwitch = "258"
    BasisFeatureDate = "259"
    BasisFeaturePrice = "260"
    MDReqID = "262"
    SubscriptionRequestType = "263"
    MarketDepth = "264"
    MDUpdateType = "265"
    AggregatedBook = "266"
    NoMDEntryTypes = "267"
    NoMDEntries = "268"
    MDEntryType = "269"
    MDEntryPx = "270"
    MDEntrySize = "271"
    MDEntryDate = "272"
    MDEntryTime = "273"
    TickDirection = "274"
    MDMkt = "275"
    QuoteCondition = "276"
    TradeCondition = "277"
    MDEntryID = "278"
    MDUpdateAction = "279"
    MDEntryRefID = "280"
    MDReqRejReason = "281"
    MDEntryOriginator = "282"
    LocationID = "283"
    DeskID = "284"
    DeleteReason = "285"
    OpenCloseSettlFlag = "286"
    SellerDays = "287"
    MDEntryBuyer = "288"
    MDEntrySeller = "289"
    MDEntryPositionNo = "290"
    FinancialStatus = "291"
    CorporateAction = "292"
    DefBidSize = "293"
    DefOfferSize = "294"
    NoQuoteEntries = "295"
    NoQuoteSets = "296"
    QuoteStatus = "297"
    QuoteCancelType = "298"
    QuoteEntryID = "299"
    QuoteRejectReason = "300"
    QuoteResponseLevel = "301"
    QuoteSetID = "302"
    QuoteRequestType = "303"
    TotNoQuoteEntries = "304"
    UnderlyingSecurityIDSource = "305"
    UnderlyingIssuer = "306"
    UnderlyingSecurityDesc = "307"
    UnderlyingSecurityExchange = "308"
    UnderlyingSecurityID = "309"
    UnderlyingSecurityType = "310"
    UnderlyingSymbol = "311"
    UnderlyingSymbolSfx = "312"
    UnderlyingMaturityMonthYear = "313"
    UnderlyingMaturityDay = "314"
    UnderlyingPutOrCall = "315"
    UnderlyingStrikePrice = "316"
    UnderlyingOptAttribute = "317"
    UnderlyingCurrency = "318"
    RatioQty = "319"
    SecurityReqID = "320"
    SecurityRequestType = "321"
    SecurityResponseID = "322"
    SecurityResponseType = "323"
    SecurityStatusReqID = "324"
    UnsolicitedIndicator = "325"
    SecurityTradingStatus = "326"
    HaltReason = "327"
    InViewOfCommon = "328"
    DueToRelated = "329"
    BuyVolume = "330"
    SellVolume = "331"
    HighPx = "332"
    LowPx = "333"
    Adjustment = "334"
    TradSesReqID = "335"
    TradingSessionID = "336"
    ContraTrader = "337"
    TradSesMethod = "338"
    TradSesMode = "339"
    TradSesStatus = "340"
    TradSesStartTime = "341"
    TradSesOpenTime = "342"
    TradSesPreCloseTime = "343"
    TradSesCloseTime = "344"
    TradSesEndTime = "345"
    NumberOfOrders = "346"
    MessageEncoding = "347"
    EncodedIssuerLen = "348"
    EncodedIssuer = "349"
    EncodedSecurityDescLen = "350"
    EncodedSecurityDesc = "351"
    EncodedListExecInstLen = "352"
    EncodedListExecInst = "353"
    EncodedTextLen = "354"
    EncodedText = "355"
    EncodedSubjectLen = "356"
    EncodedSubject = "357"
    EncodedHeadlineLen = "358"
    EncodedHeadline = "359"
    EncodedAllocTextLen = "360"
    EncodedAllocText = "361"
    EncodedUnderlyingIssuerLen = "362"
    EncodedUnderlyingIssuer = "363"
    EncodedUnderlyingSecurityDescLen = "364"
    EncodedUnderlyingSecurityDesc = "365"
    AllocPrice = "366"
    QuoteSetValidUntilTime = "367"
    QuoteEntryRejectReason = "368"
    LastMsgSeqNumProcessed = "369"
    OnBehalfOfSendingTime = "370"
    RefTagID = "371"
    RefMsgType = "372"
    SessionRejectReason = "373"
    BidRequestTransType = "374"
    ContraBroker = "375"
    ComplianceID = "376"
    SolicitedFlag = "377"
    ExecRestatementReason = "378"
    BusinessRejectRefID = "379"
    BusinessRejectReason = "380"
    GrossTradeAmt = "381"
    NoContraBrokers = "382"
    MaxMessageSize = "383"
    NoMsgTypes = "384"
    MsgDirection = "385"
    NoTradingSessions = "386"
    TotalVolumeTraded = "387"
    DiscretionInst = "388"
    DiscretionOffsetValue = "389"
    BidID = "390"
    ClientBidID = "391"
    ListName = "392"
    TotNoRelatedSym = "393"
    BidType = "394"
    NumTickets = "395"
    SideValue1 = "396"
    SideValue2 = "397"
    NoBidDescriptors = "398"
    BidDescriptorType = "399"
    BidDescriptor = "400"
    SideValueInd = "401"
    LiquidityPctLow = "402"
    LiquidityPctHigh = "403"
    LiquidityValue = "404"
    EFPTrackingError = "405"
    FairValue = "406"
    OutsideIndexPct = "407"
    ValueOfFutures = "408"
    LiquidityIndType = "409"
    WtAverageLiquidity = "410"
    ExchangeForPhysical = "411"
    OutMainCntryUIndex = "412"
    CrossPercent = "413"
    ProgRptReqs = "414"
    ProgPeriodInterval = "415"
    IncTaxInd = "416"
    NumBidders = "417"
    BidTradeType = "418"
    BasisPxType = "419"
    NoBidComponents = "420"
    Country = "421"
    TotNoStrikes = "422"
    PriceType = "423"
    DayOrderQty = "424"
    DayCumQty = "425"
    DayAvgPx = "426"
    GTBookingInst = "427"
    NoStrikes = "428"
    ListStatusType = "429"
    NetGrossInd = "430"
    ListOrderStatus = "431"
    ExpireDate = "432"
    ListExecInstType = "433"
    CxlRejResponseTo = "434"
    UnderlyingCouponRate = "435"
    UnderlyingContractMultiplier = "436"
    ContraTradeQty = "437"
    ContraTradeTime = "438"
    ClearingFirm = "439"
    ClearingAccount = "440"
    LiquidityNumSecurities = "441"
    MultiLegReportingType = "442"
    StrikeTime = "443"
    ListStatusText = "444"
    EncodedListStatusTextLen = "445"
    EncodedListStatusText = "446"
    PartyIDSource = "447"
    PartyID = "448"
    TotalVolumeTradedDate = "449"
    NetChgPrevDay = "451"
    PartyRole = "452"
    NoPartyIDs = "453"
    NoSecurityAltID = "454"
    SecurityAltID = "455"
    SecurityAltIDSource = "456"
    NoUnderlyingSecurityAltID = "457"
    UnderlyingSecurityAltID = "458"
    UnderlyingSecurityAltIDSource = "459"
    Product = "460"
    CFICode = "461"
    UnderlyingProduct = "462"
    UnderlyingCFICode = "463"
    TestMessageIndicator = "464"
    QuantityType = "465"
    BookingRefID = "466"
    IndividualAllocID = "467"
    RoundingDirection = "468"
    RoundingModulus = "469"
    CountryOfIssue = "470"
    StateOrProvinceOfIssue = "471"
    LocaleOfIssue = "472"
    NoRegistDtls = "473"
    MailingDtls = "474"
    InvestorCountryOfResidence = "475"
    PaymentRef = "476"
    DistribPaymentMethod = "477"
    CashDistribCurr = "478"
    CommCurrency = "479"
    CancellationRights = "480"
    MoneyLaunderingStatus = "481"
    MailingInst = "482"
    TransBkdTime = "483"
    ExecPriceType = "484"
    ExecPriceAdjustment = "485"
    DateOfBirth = "486"
    TradeReportTransType = "487"
    CardHolderName = "488"
    CardNumber = "489"
    CardExpDate = "490"
    CardIssNum = "491"
    PaymentMethod = "492"
    RegistAcctType = "493"
    Designation = "494"
    TaxAdvantageType = "495"
    RegistRejReasonText = "496"
    FundRenewWaiv = "497"
    CashDistribAgentName = "498"
    CashDistribAgentCode = "499"
    CashDistribAgentAcctNumber = "500"
    CashDistribPayRef = "501"
    CashDistribAgentAcctName = "502"
    CardStartDate = "503"
    PaymentDate = "504"
    PaymentRemitterID = "505"
    RegistStatus = "506"
    RegistRejReasonCode = "507"
    RegistRefID = "508"
    RegistDtls = "509"
    NoDistribInsts = "510"
    RegistEmail = "511"
    DistribPercentage = "512"
    RegistID = "513"
    RegistTransType = "514"
    ExecValuationPoint = "515"
    OrderPercent = "516"
    OwnershipType = "517"
    NoContAmts = "518"
    ContAmtType = "519"
    ContAmtValue = "520"
    ContAmtCurr = "521"
    OwnerType = "522"
    PartySubID = "523"
    NestedPartyID = "524"
    NestedPartyIDSource = "525"
    SecondaryClOrdID = "526"
    SecondaryExecID = "527"
    OrderCapacity = "528"
    OrderRestrictions = "529"
    MassCancelRequestType = "530"
    MassCancelResponse = "531"
    MassCancelRejectReason = "532"
    TotalAffectedOrders = "533"
    NoAffectedOrders = "534"
    AffectedOrderID = "535"
    AffectedSecondaryOrderID = "536"
    QuoteType = "537"
    NestedPartyRole = "538"
    NoNestedPartyIDs = "539"
    TotalAccruedInterestAmt = "540"
    MaturityDate = "541"
    UnderlyingMaturityDate = "542"
    InstrRegistry = "543"
    CashMargin = "544"
    NestedPartySubID = "545"
    Scope = "546"
    MDImplicitDelete = "547"
    CrossID = "548"
    CrossType = "549"
    CrossPrioritization = "550"
    OrigCrossID = "551"
    NoSides = "552"
    Username = "553"
    Password = "554"
    NoLegs = "555"
    LegCurrency = "556"
    TotNoSecurityTypes = "557"
    NoSecurityTypes = "558"
    SecurityListRequestType = "559"
    SecurityRequestResult = "560"
    RoundLot = "561"
    MinTradeVol = "562"
    MultiLegRptTypeReq = "563"
    LegPositionEffect = "564"
    LegCoveredOrUncovered = "565"
    LegPrice = "566"
    TradSesStatusRejReason = "567"
    TradeRequestID = "568"
    TradeRequestType = "569"
    PreviouslyReported = "570"
    TradeReportID = "571"
    TradeReportRefID = "572"
    MatchStatus = "573"
    MatchType = "574"
    OddLot = "575"
    NoClearingInstructions = "576"
    ClearingInstruction = "577"
    TradeInputSource = "578"
    TradeInputDevice = "579"
    NoDates = "580"
    AccountType = "581"
    CustOrderCapacity = "582"
    ClOrdLinkID = "583"
    MassStatusReqID = "584"
    MassStatusReqType = "585"
    OrigOrdModTime = "586"
    LegSettlType = "587"
    LegSettlDate = "588"
    DayBookingInst = "589"
    BookingUnit = "590"
    PreallocMethod = "591"
    UnderlyingCountryOfIssue = "592"
    UnderlyingStateOrProvinceOfIssue = "593"
    UnderlyingLocaleOfIssue = "594"
    UnderlyingInstrRegistry = "595"
    LegCountryOfIssue = "596"
    LegStateOrProvinceOfIssue = "597"
    LegLocaleOfIssue = "598"
    LegInstrRegistry = "599"
    LegSymbol = "600"
    LegSymbolSfx = "601"
    LegSecurityID = "602"
    LegSecurityIDSource = "603"
    NoLegSecurityAltID = "604"
    LegSecurityAltID = "605"
    LegSecurityAltIDSource = "606"
    LegProduct = "607"
    LegCFICode = "608"
    LegSecurityType = "609"
    LegMaturityMonthYear = "610"
    LegMaturityDate = "611"
    LegStrikePrice = "612"
    LegOptAttribute = "613"
    LegContractMultiplier = "614"
    LegCouponRate = "615"
    LegSecurityExchange = "616"
    LegIssuer = "617"
    EncodedLegIssuerLen = "618"
    EncodedLegIssuer = "619"
    LegSecurityDesc = "620"
    EncodedLegSecurityDescLen = "621"
    EncodedLegSecurityDesc = "622"
    LegRatioQty = "623"
    LegSide = "624"
    TradingSessionSubID = "625"
    AllocType = "626"
    NoHops = "627"
    HopCompID = "628"
    HopSendingTime = "629"
    HopRefID = "630"
    MidPx = "631"
    BidYield = "632"
    MidYield = "633"
    OfferYield = "634"
    ClearingFeeIndicator = "635"
    WorkingIndicator = "636"
    LegLastPx = "637"
    PriorityIndicator = "638"
    PriceImprovement = "639"
    Price2 = "640"
    LastForwardPoints2 = "641"
    BidForwardPoints2 = "642"
    OfferForwardPoints2 = "643"
    RFQReqID = "644"
    MktBidPx = "645"
    MktOfferPx = "646"
    MinBidSize = "647"
    MinOfferSize = "648"
    QuoteStatusReqID = "649"
    LegalConfirm = "650"
    UnderlyingLastPx = "651"
    UnderlyingLastQty = "652"
    SecDefStatus = "653"
    LegRefID = "654"
    ContraLegRefID = "655"
    SettlCurrBidFxRate = "656"
    SettlCurrOfferFxRate = "657"
    QuoteRequestRejectReason = "658"
    SideComplianceID = "659"
    AcctIDSource = "660"
    AllocAcctIDSource = "661"
    BenchmarkPrice = "662"
    BenchmarkPriceType = "663"
    ConfirmID = "664"
    ConfirmStatus = "665"
    ConfirmTransType = "666"
    ContractSettlMonth = "667"
    DeliveryForm = "668"
    LastParPx = "669"
    NoLegAllocs = "670"
    LegAllocAccount = "671"
    LegIndividualAllocID = "672"
    LegAllocQty = "673"
    LegAllocAcctIDSource = "674"
    LegSettlCurrency = "675"
    LegBenchmarkCurveCurrency = "676"
    LegBenchmarkCurveName = "677"
    LegBenchmarkCurvePoint = "678"
    LegBenchmarkPrice = "679"
    LegBenchmarkPriceType = "680"
    LegBidPx = "681"
    LegIOIQty = "682"
    NoLegStipulations = "683"
    LegOfferPx = "684"
    LegOrderQty = "685"
    LegPriceType = "686"
    LegQty = "687"
    LegStipulationType = "688"
    LegStipulationValue = "689"
    LegSwapType = "690"
    Pool = "691"
    QuotePriceType = "692"
    QuoteRespID = "693"
    QuoteRespType = "694"
    QuoteQualifier = "695"
    YieldRedemptionDate = "696"
    YieldRedemptionPrice = "697"
    YieldRedemptionPriceType = "698"
    BenchmarkSecurityID = "699"
    ReversalIndicator = "700"
    YieldCalcDate = "701"
    NoPositions = "702"
    PosType = "703"
    LongQty = "704"
    ShortQty = "705"
    PosQtyStatus = "706"
    PosAmtType = "707"
    PosAmt = "708"
    PosTransType = "709"
    PosReqID = "710"
    NoUnderlyings = "711"
    PosMaintAction = "712"
    OrigPosReqRefID = "713"
    PosMaintRptRefID = "714"
    ClearingBusinessDate = "715"
    SettlSessID = "716"
    SettlSessSubID = "717"
    AdjustmentType = "718"
    ContraryInstructionIndicator = "719"
    PriorSpreadIndicator = "720"
    PosMaintRptID = "721"
    PosMaintStatus = "722"
    PosMaintResult = "723"
    PosReqType = "724"
    ResponseTransportType = "725"
    ResponseDestination = "726"
    TotalNumPosReports = "727"
    PosReqResult = "728"
    PosReqStatus = "729"
    SettlPrice = "730"
    SettlPriceType = "731"
    UnderlyingSettlPrice = "732"
    UnderlyingSettlPriceType = "733"
    PriorSettlPrice = "734"
    NoQuoteQualifiers = "735"
    AllocSettlCurrency = "736"
    AllocSettlCurrAmt = "737"
    InterestAtMaturity = "738"
    LegDatedDate = "739"
    LegPool = "740"
    AllocInterestAtMaturity = "741"
    AllocAccruedInterestAmt = "742"
    DeliveryDate = "743"
    AssignmentMethod = "744"
    AssignmentUnit = "745"
    OpenInterest = "746"
    ExerciseMethod = "747"
    TotNumTradeReports = "748"
    TradeRequestResult = "749"
    TradeRequestStatus = "750"
    TradeReportRejectReason = "751"
    SideMultiLegReportingType = "752"
    NoPosAmt = "753"
    AutoAcceptIndicator = "754"
    AllocReportID = "755"
    NoNested2PartyIDs = "756"
    Nested2PartyID = "757"
    Nested2PartyIDSource = "758"
    Nested2PartyRole = "759"
    Nested2PartySubID = "760"
    BenchmarkSecurityIDSource = "761"
    SecuritySubType = "762"
    UnderlyingSecuritySubType = "763"
    LegSecuritySubType = "764"
    AllowableOneSidednessPct = "765"
    AllowableOneSidednessValue = "766"
    AllowableOneSidednessCurr = "767"
    NoTrdRegTimestamps = "768"
    TrdRegTimestamp = "769"
    TrdRegTimestampType = "770"
    TrdRegTimestampOrigin = "771"
    ConfirmRefID = "772"
    ConfirmType = "773"
    ConfirmRejReason = "774"
    BookingType = "775"
    IndividualAllocRejCode = "776"
    SettlInstMsgID = "777"
    NoSettlInst = "778"
    LastUpdateTime = "779"
    AllocSettlInstType = "780"
    NoSettlPartyIDs = "781"
    SettlPartyID = "782"
    SettlPartyIDSource = "783"
    SettlPartyRole = "784"
    SettlPartySubID = "785"
    SettlPartySubIDType = "786"
    DlvyInstType = "787"
    TerminationType = "788"
    NextExpectedMsgSeqNum = "789"
    OrdStatusReqID = "790"
    SettlInstReqID = "791"
    SettlInstReqRejCode = "792"
    SecondaryAllocID = "793"
    AllocReportType = "794"
    AllocReportRefID = "795"
    AllocCancReplaceReason = "796"
    CopyMsgIndicator = "797"
    AllocAccountType = "798"
    OrderAvgPx = "799"
    OrderBookingQty = "800"
    NoSettlPartySubIDs = "801"
    NoPartySubIDs = "802"
    PartySubIDType = "803"
    NoNestedPartySubIDs = "804"
    NestedPartySubIDType = "805"
    NoNested2PartySubIDs = "806"
    Nested2PartySubIDType = "807"
    AllocIntermedReqType = "808"
    UnderlyingPx = "810"
    PriceDelta = "811"
    ApplQueueMax = "812"
    ApplQueueDepth = "813"
    ApplQueueResolution = "814"
    ApplQueueAction = "815"
    NoAltMDSource = "816"
    AltMDSourceID = "817"
    SecondaryTradeReportID = "818"
    AvgPxIndicator = "819"
    TradeLinkID = "820"
    OrderInputDevice = "821"
    UnderlyingTradingSessionID = "822"
    UnderlyingTradingSessionSubID = "823"
    TradeLegRefID = "824"
    ExchangeRule = "825"
    TradeAllocIndicator = "826"
    ExpirationCycle = "827"
    TrdType = "828"
    TrdSubType = "829"
    TransferReason = "830"
    AsgnReqID = "831"
    TotNumAssignmentReports = "832"
    AsgnRptID = "833"
    ThresholdAmount = "834"
    PegMoveType = "835"
    PegOffsetType = "836"
    PegLimitType = "837"
    PegRoundDirection = "838"
    PeggedPrice = "839"
    PegScope = "840"
    DiscretionMoveType = "841"
    DiscretionOffsetType = "842"
    DiscretionLimitType = "843"
    DiscretionRoundDirection = "844"
    DiscretionPrice = "845"
    DiscretionScope = "846"
    TargetStrategy = "847"
    TargetStrategyParameters = "848"
    ParticipationRate = "849"
    TargetStrategyPerformance = "850"
    LastLiquidityInd = "851"
    PublishTrdIndicator = "852"
    ShortSaleReason = "853"
    QtyType = "854"
    SecondaryTrdType = "855"
    TradeReportType = "856"
    AllocNoOrdersType = "857"
    SharedCommission = "858"
    ConfirmReqID = "859"
    AvgParPx = "860"
    ReportedPx = "861"
    NoCapacities = "862"
    OrderCapacityQty = "863"
    NoEvents = "864"
    EventType = "865"
    EventDate = "866"
    EventPx = "867"
    EventText = "868"
    PctAtRisk = "869"
    NoInstrAttrib = "870"
    InstrAttribType = "871"
    InstrAttribValue = "872"
    DatedDate = "873"
    InterestAccrualDate = "874"
    CPProgram = "875"
    CPRegType = "876"
    UnderlyingCPProgram = "877"
    UnderlyingCPRegType = "878"
    UnderlyingQty = "879"
    TrdMatchID = "880"
    SecondaryTradeReportRefID = "881"
    UnderlyingDirtyPrice = "882"
    UnderlyingEndPrice = "883"
    UnderlyingStartValue = "884"
    UnderlyingCurrentValue = "885"
    UnderlyingEndValue = "886"
    NoUnderlyingStips = "887"
    UnderlyingStipType = "888"
    UnderlyingStipValue = "889"
    MaturityNetMoney = "890"
    MiscFeeBasis = "891"
    TotNoAllocs = "892"
    LastFragment = "893"
    CollReqID = "894"
    CollAsgnReason = "895"
    CollInquiryQualifier = "896"
    NoTrades = "897"
    MarginRatio = "898"
    MarginExcess = "899"
    TotalNetValue = "900"
    CashOutstanding = "901"
    CollAsgnID = "902"
    CollAsgnTransType = "903"
    CollRespID = "904"
    CollAsgnRespType = "905"
    CollAsgnRejectReason = "906"
    CollAsgnRefID = "907"
    CollRptID = "908"
    CollInquiryID = "909"
    CollStatus = "910"
    TotNumReports = "911"
    LastRptRequested = "912"
    AgreementDesc = "913"
    AgreementID = "914"
    AgreementDate = "915"
    StartDate = "916"
    EndDate = "917"
    AgreementCurrency = "918"
    DeliveryType = "919"
    EndAccruedInterestAmt = "920"
    StartCash = "921"
    EndCash = "922"
    UserRequestID = "923"
    UserRequestType = "924"
    NewPassword = "925"
    UserStatus = "926"
    UserStatusText = "927"
    StatusValue = "928"
    StatusText = "929"
    RefCompID = "930"
    RefSubID = "931"
    NetworkResponseID = "932"
    NetworkRequestID = "933"
    LastNetworkResponseID = "934"
    NetworkRequestType = "935"
    NoCompIDs = "936"
    NetworkStatusResponseType = "937"
    NoCollInquiryQualifier = "938"
    TrdRptStatus = "939"
    AffirmStatus = "940"
    UnderlyingStrikeCurrency = "941"
    LegStrikeCurrency = "942"
    TimeBracket = "943"
    CollAction = "944"
    CollInquiryStatus = "945"
    CollInquiryResult = "946"
    StrikeCurrency = "947"
    NoNested3PartyIDs = "948"
    Nested3PartyID = "949"
    Nested3PartyIDSource = "950"
    Nested3PartyRole = "951"
    NoNested3PartySubIDs = "952"
    Nested3PartySubID = "953"
    Nested3PartySubIDType = "954"
    LegContractSettlMonth = "955"
    LegInterestAccrualDate = "956"



import logging

from asyncfix import FTag
from asyncfix.message import MessageDirection


class FIXSession:
    def __init__(self, key, target_comp_id, sender_comp_id):
        self.key = key
        self.sender_comp_id = sender_comp_id
        self.target_comp_id = target_comp_id

        self.snd_seq_num = None
        self.messages = None
        self.next_expected_msg_seq_num = None

        self.reset_msgs()

    def reset_msgs(self):
        self.snd_seq_num = 0
        self.next_expected_msg_seq_num = 1
        self.messages = {MessageDirection.OUTBOUND: {}, MessageDirection.INBOUND: {}}

    def validate_comp_ids(self, target_comp_id, sender_comp_id):
        return (
            self.sender_comp_id == sender_comp_id
            and self.target_comp_id == target_comp_id
        )

    def allocate_snd_seq_no(self):
        self.snd_seq_num += 1
        return str(self.snd_seq_num)

    def validate_recv_seq_no(self, seq_no):
        if self.next_expected_msg_seq_num < int(seq_no):
            logging.warning(
                "SeqNum from client unexpected (Rcvd: %s Expected: %s)"
                % (seq_no, self.next_expected_msg_seq_num)
            )
            return (False, self.next_expected_msg_seq_num)
        else:
            return (True, seq_no)

    def reset_seq_num(self):
        self.snd_seq_num = 0
        self.next_expected_msg_seq_num = 1

    def set_recv_seq_no(self, seq_no):
        # if self.nextExpectedMsgSeqNum != int(seqNo):
        #     logging.warning("SeqNum from client unexpected (Rcvd: %s Expected: %s)"
        #           % (seqNo, self.nextExpectedMsgSeqNum))
        self.next_expected_msg_seq_num = int(seq_no) + 1

    def persist_msg(self, msg, direction):
        seqNo = msg[FTag.MsgSeqNum]
        self.messages[direction][seqNo] = msg



import pytest

from asyncfix import FMsg, FTag


def test_tags():
    assert FTag.Account == "1"
    assert FTag.Account == 1
    assert str(FTag.Account) == "1"
    assert f"{FTag.Account}" == "1"
    assert FTag("1") == "1"
    assert FTag("1") == FTag.Account
    assert FTag(FTag.Account) == FTag.Account
    assert FTag.Account in FTag
    assert "1" in FTag
    assert 1 in FTag
    assert 1102938109238 not in FTag
    assert "asdad" not in FTag
    assert hash(FTag.Account) == hash("1")


def test_msgtype():
    assert FMsg.HEARTBEAT == "0"
    assert FMsg.HEARTBEAT != 0
    assert str(FMsg.HEARTBEAT) == "0"
    assert f"{FMsg.HEARTBEAT}" == "0"
    assert FMsg("0") == FMsg.HEARTBEAT
    assert "0" in FMsg
    assert 0 not in FMsg
    assert hash(FMsg.HEARTBEAT) == hash("0")
    assert FMsg.HEARTBEAT in FMsg



import datetime
import importlib
import unittest
from unittest.mock import Mock, patch

import pytest

from asyncfix import FMsg, FTag
from asyncfix.codec import Codec
from asyncfix.message import (
    FIXContainer,
    FIXMessage,
    RepeatingTagError,
    TagNotFoundError,
    UnmappedRepeatedGrpError,
)
from asyncfix.protocol import FIXProtocol44


class FakeDate(datetime.datetime):
    @classmethod
    def utcnow(cls):
        return cls(2015, 6, 19, 11, 8, 54)


@pytest.fixture
def fix_session():
    mock_session = Mock()
    mock_session.sender_comp_id = "sender"
    mock_session.target_comp_id = "target"
    mock_session.allocate_snd_seq_no.return_value = 1
    return mock_session


def test_decode():
    protocol = FIXProtocol44()
    codec = Codec(protocol)
    inMsg = (
        b"8=FIX.4.4\x019=817\x0135=J\x0134=953\x0149=FIX_ALAUDIT\x0156=BFUT_ALAUDIT\x0143=N\x0152=20150615-09:21:42.459\x0170=00000002664ASLO1001\x01626=2\x0110626=5\x0171=0\x0160=20150615-10:21:42\x01857=1\x0173=1\x0111=00000006321ORLO1\x0138=100.0\x01800=100.0\x01124=1\x0132=100.0\x0117=00000009758TRLO1\x0131=484.50\x0154=2\x0153=100.0\x0155=FTI\x01207=XEUE\x01454=1\x01455=EOM5\x01456=A\x01200=201506\x01541=20150619\x01461=FXXXXX\x016=484.50\x0174=2\x0175=20150615\x0178=2\x0179=TEST123\x0130009=12345\x01467=00000014901CALO1001\x019520=00000014898CALO1\x0180=33.0\x01366=484.50\x0181=0\x01153=484.50\x0110626=5\x0179=TEST124\x0130009=12345\x01467=00000014903CALO1001\x019520=00000014899CALO1\x0180=67.0\x01366=484.50\x0181=0\x01153=484.50\x0110626=5\x01453=3\x01448=TEST1\x01447=D\x01452=3\x01802=2\x01523=12345\x01803=3\x01523=TEST1\x01803=19\x01448=TEST1WA\x01447=D\x01452=38\x01802=4\x01523=Test1"
        b" Wait\x01803=10\x01523="
        b" \x01803=26\x01523=\x01803=3\x01523=TestWaCRF2\x01803=28\x01448=hagap\x01447=D\x01452=11\x01802=2\x01523=GB\x01803=25\x01523=BarCapFutures.FETService\x01803=24\x0110=033\x01"
    )
    msg, remaining = codec.decode(inMsg)

    exp_str = "8=FIX.4.4|9=817|35=J|34=953|49=FIX_ALAUDIT|56=BFUT_ALAUDIT|43=N|52=20150615-09:21:42.459|70=00000002664ASLO1001|626=2|10626=#err#|71=0|60=20150615-10:21:42|857=1|73=1=>[11=00000006321ORLO1|38=100.0|800=100.0]|124=1=>[32=100.0|17=00000009758TRLO1|31=484.50]|54=2|53=100.0|55=FTI|207=XEUE|454=1=>[455=EOM5|456=A]|200=201506|541=20150619|461=FXXXXX|6=484.50|74=2|75=20150615|78=1=>[79=TEST123]|30009=#err#|467=#err#|9520=#err#|80=#err#|366=#err#|81=#err#|153=#err#|79=TEST124|453=3=>[448=TEST1|447=D|452=3|802=2=>[523=12345|803=3, 523=TEST1|803=19], 448=TEST1WA|447=D|452=38|802=4=>[523=Test1 Wait|803=10, 523= |803=26, 523=|803=3, 523=TestWaCRF2|803=28], 448=hagap|447=D|452=11|802=2=>[523=GB|803=25, 523=BarCapFutures.FETService|803=24]]|10=033"  # noqa
    assert exp_str == str(msg)


def test_encode(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    msg = FIXMessage(codec.protocol.msgtype.NEWORDERSINGLE)
    msg.set(codec.protocol.fixtags.Price, "123.45")
    msg.set(codec.protocol.fixtags.OrderQty, 9876)
    msg.set(codec.protocol.fixtags.Symbol, "VOD.L")
    msg.set(codec.protocol.fixtags.SecurityID, "GB00BH4HKS39")
    msg.set(codec.protocol.fixtags.SecurityIDSource, "4")
    msg.set(codec.protocol.fixtags.Account, "TEST")
    msg.set(codec.protocol.fixtags.HandlInst, "1")
    msg.set(codec.protocol.fixtags.ExDestination, "XLON")
    msg.set(codec.protocol.fixtags.Side, 1)
    msg.set(codec.protocol.fixtags.ClOrdID, "abcdefg")
    msg.set(codec.protocol.fixtags.Currency, "GBP")

    rptgrp1 = FIXContainer()
    rptgrp1.set("611", "aaa")
    rptgrp1.set("612", "bbb")
    rptgrp1.set("613", "ccc")

    msg.add_group("444", rptgrp1, 0)

    rptgrp2 = FIXContainer()
    rptgrp2.set("611", "zzz")
    rptgrp2.set("612", "yyy")
    rptgrp2.set("613", "xxx")
    msg.add_group("444", rptgrp2, 1)

    with patch("asyncfix.codec.datetime", FakeDate) as mock_date:
        result = codec.encode(msg, fix_session)
    expected = "8=FIX.4.4\x019=201\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20150619-11:08:54.000\x0144=123.45\x0138=9876\x0155=VOD.L\x0148=GB00BH4HKS39\x0122=4\x011=TEST\x0121=1\x01100=XLON\x0154=1\x0111=abcdefg\x0115=GBP\x01444=2\x01611=aaa\x01612=bbb\x01613=ccc\x01611=zzz\x01612=yyy\x01613=xxx\x0110=255\x01"
    assert expected == result


def test_decode_invalid_checksum(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    msg = FIXMessage(codec.protocol.msgtype.NEWORDERSINGLE)
    msg.set(codec.protocol.fixtags.Price, "123.45")
    msg.set(codec.protocol.fixtags.OrderQty, 9876)
    msg.set(codec.protocol.fixtags.Symbol, "VOD.L")
    protocol = FIXProtocol44()
    codec = Codec(protocol)
    # enc_msg = codec.encode(msg, fix_session)
    # print(repr(enc_msg))
    # assert False, enc_msg

    enc_msg = b"8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=110\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)

    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="invalid checksum tag"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_valid(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    msg = FIXMessage(codec.protocol.msgtype.NEWORDERSINGLE)
    msg.set(codec.protocol.fixtags.Price, "123.45")
    msg.set(codec.protocol.fixtags.OrderQty, 9876)
    msg.set(codec.protocol.fixtags.Symbol, "VOD.L")
    protocol = FIXProtocol44()
    codec = Codec(protocol)
    # enc_msg = codec.encode(msg, fix_session)
    # print(repr(enc_msg))
    # assert False, enc_msg

    enc_msg = b"8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg, silent=False)

    assert isinstance(msg, FIXMessage)
    assert msg[8] == "FIX.4.4"
    assert msg.msg_type == FMsg.NEWORDERSINGLE
    assert msg[FTag.Price] == "123.45"
    assert msg[FTag.OrderQty] == "9876"
    assert msg[FTag.Symbol] == "VOD.L"
    assert parsed_len == len(enc_msg)


def test_decode_invalid_no_fix(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"my_some string without any "  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="no fix header"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_invalid_with_added_fix(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"somejunk\n8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert isinstance(msg, FIXMessage)
    assert parsed_len == len(enc_msg)

    msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_invalid_junk_with_incomplete_fix(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"somejunk\n8=FIX.4.4\x019=82\x0135=D"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    # assert isinstance(msg, FIXMessage)
    assert msg is None
    assert parsed_len == len("somejunk\n")
    assert enc_msg[parsed_len:].startswith(b"8=FIX")

    with pytest.raises(AssertionError, match="incomplete message"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_invalid_2nd_fix_msg(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"somejunk\n8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x018=FIX.4.4\x019=82\x0135=D"  # noqa
    # msg, parsed_len = codec.decode(enc_msg)
    msg, parsed_len = codec.decode(enc_msg, silent=False)
    assert isinstance(msg, FIXMessage)
    assert parsed_len == len(
        b"somejunk\n8=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"
    )
    assert enc_msg[parsed_len:].startswith(b"8=FIX")


def test_decode_invalid_start_fix_msg(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"35=D\x0149=sender\x0156=target\x0134=1\x018=FIX.4.4\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert isinstance(msg, FIXMessage)
    assert parsed_len == len(enc_msg)

    msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_groups(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    msg_in = FIXMessage(codec.protocol.msgtype.NEWORDERSINGLE)
    msg_in.set(codec.protocol.fixtags.Price, "123.45")
    msg_in.set(codec.protocol.fixtags.OrderQty, 9876)
    msg_in.set(codec.protocol.fixtags.Symbol, "VOD.L")

    rptgrp1 = FIXContainer()
    rptgrp1.set(FTag.SecurityAltID, "abc")
    rptgrp1.set(FTag.SecurityAltIDSource, "bbb")
    msg_in.add_group(FTag.NoSecurityAltID, rptgrp1)

    rptgrp2 = FIXContainer()
    rptgrp2.set(FTag.SecurityAltID, "zzz")
    rptgrp2.set(FTag.SecurityAltIDSource, "xxx")
    msg_in.add_group(FTag.NoSecurityAltID, rptgrp2)

    g = FIXContainer()
    g.set("20323", "1")
    g.set("20324", "3")
    msg_in.add_group("20228", g)

    g = FIXContainer()
    g.set("20323", "1")
    g.set("20324", "3")
    msg_in.add_group("20228", g)

    protocol = FIXProtocol44()
    codec = Codec(protocol)
    enc_msg = codec.encode(msg_in, fix_session).encode()
    # print(repr(enc_msg))
    # assert False, enc_msg

    msg_out, parsed_len = codec.decode(enc_msg)

    assert isinstance(msg_out, FIXMessage)
    assert msg_out[8] == "FIX.4.4"
    assert msg_out.msg_type == FMsg.NEWORDERSINGLE
    assert msg_out[FTag.Price] == "123.45"
    assert msg_out[FTag.OrderQty] == "9876"
    assert msg_out[FTag.Symbol] == "VOD.L"

    # print(repr(msg_out))
    # assert False, enc_msg

    g = msg_out.get_group_list(FTag.NoSecurityAltID)
    assert g
    assert isinstance(g, list)
    assert len(g) == 2
    assert g[0][FTag.SecurityAltID] == "abc"
    assert g[0][FTag.SecurityAltIDSource] == "bbb"
    assert g[1][FTag.SecurityAltID] == "zzz"
    assert g[1][FTag.SecurityAltIDSource] == "xxx"

    with pytest.raises(RepeatingTagError, match="tag=.*was repeated"):
        msg_out["20323"]

    with pytest.raises(
        UnmappedRepeatedGrpError,
        match="tag exists, but it does not belong to any group",
    ):
        g = msg_out.get_group_list("20228")

    with pytest.raises(TagNotFoundError, match="missing tag group tag="):
        g = msg_out.get_group_list("129012099")


def test_decode_protocol_mismatch(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"8=FIX.4.2\x019=82\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138=9876\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="protocol beginstring mismatch"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_body_len_split_err(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    # this is an error, but possibly caused by socket buffer not loaded
    #   expected to be appended by good data later
    enc_msg = b"8=FIX.4.4\x0199\x0199=2\x01"
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="BodyLength split error "):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_bad_second_tag(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    # this is an error, but possibly caused by socket buffer not loaded
    #   expected to be appended by good data later
    enc_msg = b"8=FIX.4.4\x0135=\x0199=2\x01"
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == 19

    with pytest.raises(AssertionError, match="2nd tag must be BodyLength"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_bad_second_tag_expected_body_length(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)
    # 2nd tag always must be a bodylength
    enc_msg = b"8=FIX.4.4\x0135=8\x0199=2\x01"
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="2nd tag must be BodyLength"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_with_unicode_valid(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    msg = FIXMessage(codec.protocol.msgtype.NEWORDERSINGLE)
    msg.set(codec.protocol.fixtags.Price, "123.45")
    msg.set(codec.protocol.fixtags.OrderQty, 9876)
    msg.set(codec.protocol.fixtags.Symbol, "ЮН")
    protocol = FIXProtocol44()
    codec = Codec(protocol)
    enc_msg = codec.encode(msg, fix_session).encode("cp1251")
    # print(repr(enc_msg))
    # assert False, enc_msg

    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)

    with pytest.raises(AssertionError, match="invalid checksum tag"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_empty_tag(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x0138\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)
    with pytest.raises(AssertionError, match="incomplete tag 38"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)

    enc_msg = b"8=FIX.4.4\x019=75\x0135=D\x0149=sender\x0156=target\x0134=1\x0152=20230919-07:13:26.808\x0144=123.45\x01\x0155=VOD.L\x0110=100\x01"  # noqa
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == len(enc_msg)
    with pytest.raises(AssertionError, match="incomplete tag"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)


def test_decode_minimum_message(fix_session):
    protocol = FIXProtocol44()
    codec = Codec(protocol)

    enc_msg = b"8=FIX.4.4\x019=75\x01"
    msg, parsed_len = codec.decode(enc_msg)
    assert msg is None
    assert parsed_len == 0
    with pytest.raises(AssertionError, match="Minimum message"):
        msg, parsed_len = codec.decode(enc_msg, silent=False)



import enum


class StrEnum(enum.Enum):
    def __str__(self):
        return str(self.value)

    def __eq__(self, o):
        return self.value == str(o)

    def __hash__(self):
        return hash(self.value)


class FOrdSide(StrEnum):
    BUY = "1"
    SELL = "2"
    BUY_MINUS = "3"
    SELL_PLUS = "4"
    SELL_SHORT = "5"
    SELL_SHORT_EXEMPT = "6"
    UNDISCLOSED = "7"
    CROSS = "8"
    CROSS_SHORT = "9"
    CROSS_SHORT_EXEMPT = "A"
    AS_DEFINED = "B"
    OPPOSITE = "C"
    SUBSCRIBE = "D"
    REDEEM = "E"
    LEND = "F"
    BORROW = "G"


class FOrdType(StrEnum):
    MARKET = "1"
    LIMIT = "2"
    STOP = "3"
    STOP_LIMIT = "4"
    WITH_OR_WITHOUT = "6"
    LIMIT_OR_BETTER = "7"
    LIMIT_WITH_OR_WITHOUT = "8"
    ON_BASIS = "9"
    PREVIOUSLY_QUOTED = "D"
    PREVIOUSLY_INDICATED = "E"
    FOREX_SWAP = "G"
    FUNARI = "I"
    MARKET_IF_TOUCHED = "J"
    MARKET_WITH_LEFT_OVER_AS_LIMIT = "K"
    PREVIOUS_FUND_VALUATION_POINT = "L"
    NEXT_FUND_VALUATION_POINT = "M"
    PEGGED = "P"


class FOrdStatus(StrEnum):
    CREATED = "Z"  # IMPORTANT: this one is for internal use, non FIX!
    NEW = "0"
    PARTIALLY_FILLED = "1"
    FILLED = "2"
    DONE_FOR_DAY = "3"
    CANCELED = "4"
    PENDING_CANCEL = "6"
    STOPPED = "7"
    REJECTED = "8"
    SUSPENDED = "9"
    PENDING_NEW = "A"
    CALCULATED = "B"
    EXPIRED = "C"
    ACCEPTED_FOR_BIDDING = "D"
    PENDING_REPLACE = "E"


class FExecType(StrEnum):
    NEW = "0"
    DONE_FOR_DAY = "3"
    CANCELED = "4"
    REPLACED = "5"
    PENDING_CANCEL = "6"
    STOPPED = "7"
    REJECTED = "8"
    SUSPENDED = "9"
    PENDING_NEW = "A"
    CALCULATED = "B"
    EXPIRED = "C"
    RESTATED = "D"
    PENDING_REPLACE = "E"
    TRADE = "F"
    TRADE_CORRECT = "G"
    TRADE_CANCEL = "H"
    ORDER_STATUS = "I"



from __future__ import annotations

from collections import OrderedDict
from enum import Enum
from typing import Any

from asyncfix import FMsg, FTag

from .errors import (
    DuplicatedTagError,
    FIXMessageError,
    RepeatingTagError,
    TagNotFoundError,
    UnmappedRepeatedGrpError,
)


def isclass(cl):
    try:
        return issubclass(cl, cl)
    except TypeError:
        return False


class MessageDirection(Enum):
    INBOUND = 0
    OUTBOUND = 1


class _FIXRepeatingGroupContainer:
    def __init__(self):
        self.groups: list[FIXContainer] = []

    def add_group(self, group, index):
        if index == -1:
            self.groups.append(group)
        else:
            self.groups.insert(index, group)

    def __str__(self):
        return str(len(self.groups)) + "=>" + str(self.groups)

    __repr__ = __str__


class FIXContainer(object):
    def __init__(
        self, tags: dict[str | int, [str, float, int, list[dict | FIXContainer]]] = None
    ):
        self.tags: dict[str, str | _FIXRepeatingGroupContainer] = OrderedDict()

        if tags:
            for t, v in tags.items():
                if isinstance(v, list):
                    self.set_group(t, v)
                else:
                    self.set(t, v)

    def set(self, tag: str | int, value, replace: bool = False):
        try:
            # tag also might be an FTag enum (so cast to str first)
            int(str(tag))
        except ValueError:
            raise FIXMessageError("Tags must be only integers")

        t = str(tag)

        if isclass(value):
            # Case for setting tags as errors (allow overwriting by Exception)
            value = value
        else:
            if not replace and t in self.tags:
                raise DuplicatedTagError(f"tag={t} already exists")

            value = str(value)

        self.tags[t] = value

    def get(self, tag: str | int, default=TagNotFoundError):
        result = self.tags.get(str(tag), default)
        if result is TagNotFoundError:
            raise TagNotFoundError(f"tag={tag} not found in message")
        elif result is RepeatingTagError:
            raise RepeatingTagError(
                f"tag={tag} was repeated, possible undefined repeating group or"
                " malformed fix message"
            )
        elif isinstance(result, _FIXRepeatingGroupContainer):
            raise FIXMessageError(
                "You are trying to get group as simple tag,use get_group*() methods"
            )
        return result

    def is_group(self, tag: str | int) -> bool | None:
        tag_val = self.tags.get(str(tag), None)
        if tag_val is not None:
            if isinstance(tag_val, _FIXRepeatingGroupContainer):
                return True
            else:
                return False
        else:
            return None

    def add_group(self, tag: str | int, group: FIXContainer | dict, index: int = -1):
        tag = str(tag)

        if isinstance(group, dict):
            group = FIXContainer(group)
        elif not isinstance(group, FIXContainer):
            raise FIXMessageError(f"Expected FIXContext in group, got {type(group)}")

        if tag in self:
            group_container = self.tags[tag]
            group_container.add_group(group, index)
        else:
            group_container = _FIXRepeatingGroupContainer()
            group_container.add_group(group, index)
            self.tags[tag] = group_container

    def set_group(self, tag: str | int, groups: list[dict, FIXContainer]):
        tag = str(tag)

        if tag in self:
            raise DuplicatedTagError(f"group with {tag=} already exists")

        group_container = _FIXRepeatingGroupContainer()

        for m in groups:
            if isinstance(m, dict):
                m = FIXContainer(tags=m)
            elif not isinstance(m, FIXContainer):
                raise FIXMessageError(
                    f"groups must be a list of FIXContext or dict, got {type(m)}"
                )
            group_container.add_group(m, -1)

        self.tags[tag] = group_container

    def get_group_list(self, tag: str | int) -> list[FIXContainer]:
        tag = str(tag)
        is_group = self.is_group(tag)
        if is_group is None:
            raise TagNotFoundError(f"missing tag group {tag=}")
        if not is_group:
            raise UnmappedRepeatedGrpError(
                "tag exists, but it does not belong to any group, not a group tag or"
                " missing `repeating_group` in protocol class"
            )
        return self.tags[tag].groups

    def get_group_by_tag(
        self, tag: str | int, gtag: str | int, gvalue: str
    ) -> FIXContainer:
        for group in self.get_group_list(tag):
            if gtag in group:
                if group.get(gtag) == gvalue:
                    return group
        raise TagNotFoundError(f"get_group_by_tag: {tag=} {gtag=} {gvalue=} missing")

    def get_group_by_index(self, tag: str | int, index: int) -> FIXContainer:
        g = self.get_group_list(tag)

        if index >= len(g):
            raise TagNotFoundError(
                f"get_group_by_index: index is out of range of {tag=} group"
            )

        return g[index]

    def items(self):
        return self.tags.items()

    def __getitem__(self, tag: str | int) -> Any:
        return self.get(tag)

    def __setitem__(self, tag: str | int, value):
        self.set(tag, value)

    def __delitem__(self, tag: str | int):
        del self.tags[str(tag)]

    def __contains__(self, item: str | int | dict[str, Any]):
        if isinstance(item, dict):
            assert item, "empty dict item"

            # tag member comparison
            for t, v in item.items():
                t = str(t)
                v = str(v)

                if t not in self.tags:
                    return False

                if self.is_group(t):
                    raise FIXMessageError(
                        "fix message __contains__ supports only simple tags, got group"
                        f" at tag={t}"
                    )

                if v != self.get(t):
                    return False
            return True
        else:
            return str(item) in self.tags

    def __str__(self):
        r = ""
        allTags = []
        for tag, tag_value in self.tags.items():
            if isclass(tag_value) and issubclass(tag_value, Exception):
                tag_value = "#err#"
            allTags.append("%s=%s" % (tag, tag_value))
        r += "|".join(allTags)
        return r

    def __eq__(self, other):
        # if our string representation looks the same, the objects are equivalent
        if isinstance(other, FIXContainer):
            return self.__str__() == other.__str__()
        elif isinstance(other, dict):
            ignore_tags = {
                FTag.BeginString,
                FTag.BodyLength,
                FTag.CheckSum,
                FTag.MsgType,
            }

            other_tags = set(
                [str(t) for t in other.keys() if str(t) not in ignore_tags]
            )
            self_tags = set(
                [str(t) for t in self.tags.keys() if str(t) not in ignore_tags]
            )

            if other_tags != self_tags:
                return False

            for t, v in other.items():
                if self.is_group(t):
                    raise FIXMessageError(
                        "fix message __eq__ (dict) supports only simple tags, got group"
                        f" at tag={t}"
                    )

                v = str(other[t])
                if v != self.get(t):
                    return False

            return True
        else:
            return False

    __repr__ = __str__


class FIXMessage(FIXContainer):
    def __init__(
        self,
        msg_type: str | FMsg,
        tags: dict[str | int, [str, float, int]] = None,
    ):
        self._msg_type = msg_type
        super().__init__(tags)

    @property
    def msg_type(self) -> str | FMsg:
        return self._msg_type

    @msg_type.setter
    def msg_type(self, msg_type: str | FMsg):
        self._msg_type = msg_type



from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.protocol.protocol_base import FIXProtocolBase


class FIXProtocol44(FIXProtocolBase):
    beginstring = "FIX.4.4"

    session_message_types = {
        FMsg.HEARTBEAT,
        FMsg.TESTREQUEST,
        FMsg.RESENDREQUEST,
        FMsg.REJECT,
        FMsg.SEQUENCERESET,
        FMsg.LOGOUT,
        FMsg.LOGON,
        FMsg.XMLNONFIX,
    }

    repeating_groups = {
        FTag.NoSecurityAltID: [FTag.SecurityAltID, FTag.SecurityAltIDSource],
        FTag.NoMiscFees: [
            FTag.MiscFeeAmt,
            FTag.MiscFeeCurr,
            FTag.MiscFeeType,
            FTag.MiscFeeBasis,
        ],
        FTag.NoClearingInstructions: [
            FTag.ClearingInstruction,
        ],
        FTag.NoEvents: [FTag.EventType, FTag.EventDate, FTag.EventPx, FTag.EventText],
        FTag.NoInstrAttrib: [FTag.InstrAttribType, FTag.InstrAttribValue],
        FTag.NoLegSecurityAltID: [FTag.LegSecurityAltID, FTag.LegSecurityAltIDSource],
        FTag.NoLegStipulations: [FTag.LegStipulationType, FTag.LegStipulationValue],
        FTag.NoNestedPartyIDs: [
            FTag.NestedPartyID,
            FTag.NestedPartyIDSource,
            FTag.NestedPartyRole,
            FTag.NoNestedPartySubIDs,
        ],
        FTag.NoNestedPartySubIDs: [FTag.NestedPartySubID, FTag.NestedPartySubIDType],
        FTag.NoNested2PartyIDs: [
            FTag.Nested2PartyID,
            FTag.Nested2PartyIDSource,
            FTag.Nested2PartyRole,
            FTag.NoNested2PartySubIDs,
        ],
        FTag.NoNested2PartySubIDs: [FTag.Nested2PartySubID, FTag.Nested2PartySubIDType],
        FTag.NoNested3PartyIDs: [
            FTag.Nested3PartyID,
            FTag.Nested3PartyIDSource,
            FTag.Nested3PartyRole,
            FTag.NoNested3PartySubIDs,
        ],
        FTag.NoNested3PartySubIDs: [FTag.Nested3PartySubID, FTag.Nested3PartySubIDType],
        FTag.NoPartyIDs: [
            FTag.PartyID,
            FTag.PartyIDSource,
            FTag.PartyRole,
            FTag.NoPartySubIDs,
        ],
        FTag.NoPartySubIDs: [FTag.PartySubID, FTag.PartySubIDType],
        FTag.NoPosAmt: [FTag.PosAmtType, FTag.PosAmt],
        FTag.NoPositions: [
            FTag.PosType,
            FTag.LongQty,
            FTag.ShortQty,
            FTag.PosQtyStatus,
        ],
        FTag.NoDlvyInst: [
            FTag.SettlInstSource,
            FTag.DlvyInstType,
            FTag.NoSettlPartyIDs,
        ],
        FTag.NoSettlPartyIDs: [
            FTag.SettlPartyID,
            FTag.SettlPartyIDSource,
            FTag.SettlPartyRole,
            FTag.NoSettlPartySubIDs,
        ],
        FTag.NoSettlPartySubIDs: [FTag.SettlPartySubID, FTag.SettlPartySubIDType],
        FTag.NoStipulations: [FTag.StipulationType, FTag.StipulationValue],
        FTag.NoTrdRegTimestamps: [
            FTag.TrdRegTimestamp,
            FTag.TrdRegTimestampType,
            FTag.TrdRegTimestampOrigin,
        ],
        FTag.NoUnderlyingSecurityAltID: [
            FTag.UnderlyingSecurityAltID,
            FTag.UnderlyingSecurityAltIDSource,
        ],
        FTag.NoUnderlyingStips: [FTag.UnderlyingStipType, FTag.UnderlyingStipValue],
        FTag.NoOrders: [
            FTag.ClOrdID,
            FTag.OrderID,
            FTag.SecondaryOrderID,
            FTag.SecondaryClOrdID,
            FTag.ListID,
            FTag.OrderQty,
            FTag.OrderAvgPx,
            FTag.OrderBookingQty,
        ],
        FTag.NoExecs: [
            FTag.LastQty,
            FTag.ExecID,
            FTag.SecondaryExecID,
            FTag.LastPx,
            FTag.LastParPx,
            FTag.LastCapacity,
        ],
        FTag.NoUnderlyings: [
            FTag.UnderlyingSymbol,
            FTag.UnderlyingSymbolSfx,
            FTag.UnderlyingSecurityID,
            FTag.UnderlyingSecurityIDSource,
            FTag.NoUnderlyingSecurityAltID,
            FTag.UnderlyingProduct,
            FTag.UnderlyingCFICode,
            FTag.UnderlyingSecurityType,
            FTag.UnderlyingSecuritySubType,
            FTag.UnderlyingMaturityMonthYear,
            FTag.UnderlyingMaturityDate,
            FTag.UnderlyingPutOrCall,
            FTag.UnderlyingCouponPaymentDate,
            FTag.UnderlyingIssueDate,
            FTag.UnderlyingRepoCollateralSecurityType,
            FTag.UnderlyingRepurchaseTerm,
            FTag.UnderlyingRepurchaseRate,
            FTag.UnderlyingFactor,
            FTag.UnderlyingCreditRating,
            FTag.UnderlyingInstrRegistry,
            FTag.UnderlyingCountryOfIssue,
            FTag.UnderlyingStateOrProvinceOfIssue,
            FTag.UnderlyingLocaleOfIssue,
            FTag.UnderlyingRedemptionDate,
            FTag.UnderlyingStrikePrice,
            FTag.UnderlyingStrikeCurrency,
            FTag.UnderlyingOptAttribute,
            FTag.UnderlyingContractMultiplier,
            FTag.UnderlyingCouponRate,
            FTag.UnderlyingSecurityExchange,
            FTag.UnderlyingIssuer,
            FTag.EncodedUnderlyingIssuerLen,
            FTag.EncodedUnderlyingIssuer,
            FTag.UnderlyingSecurityDesc,
            FTag.EncodedUnderlyingSecurityDescLen,
            FTag.EncodedUnderlyingSecurityDesc,
            FTag.UnderlyingCPProgram,
            FTag.UnderlyingCPRegType,
            FTag.UnderlyingCurrency,
            FTag.UnderlyingQty,
            FTag.UnderlyingPx,
            FTag.UnderlyingDirtyPrice,
            FTag.UnderlyingEndPrice,
            FTag.UnderlyingStartValue,
            FTag.UnderlyingCurrentValue,
            FTag.UnderlyingEndValue,
            FTag.NoUnderlyingStips,
        ],
        FTag.NoAllocs: [
            FTag.AllocAccount,
            FTag.AllocAcctIDSource,
            FTag.MatchStatus,
            FTag.AllocPrice,
            FTag.AllocQty,
            FTag.IndividualAllocID,
            FTag.ProcessCode,
            FTag.NoNestedPartyIDs,
            FTag.NotifyBrokerOfCredit,
            FTag.AllocHandlInst,
            FTag.AllocText,
            FTag.EncodedAllocTextLen,
            FTag.EncodedAllocText,
            FTag.Commission,
            FTag.CommType,
            FTag.CommCurrency,
            FTag.FundRenewWaiv,
            FTag.AllocAvgPx,
            FTag.AllocNetMoney,
            FTag.SettlCurrAmt,
            FTag.AllocSettlCurrAmt,
            FTag.SettlCurrency,
            FTag.AllocSettlCurrency,
            FTag.SettlCurrFxRate,
            FTag.SettlCurrFxRateCalc,
            FTag.AllocAccruedInterestAmt,
            FTag.AllocInterestAtMaturity,
            FTag.NoMiscFees,
            FTag.NoClearingInstructions,
            FTag.AllocSettlInstType,
            FTag.SettlDeliveryType,
            FTag.StandInstDbType,
            FTag.StandInstDbName,
            FTag.StandInstDbID,
            FTag.NoDlvyInst,
        ],
        FTag.NoLegs: [
            FTag.LegSymbol,
            FTag.LegSymbolSfx,
            FTag.LegSecurityID,
            FTag.LegSecurityIDSource,
            FTag.NoLegSecurityAltID,
            FTag.LegProduct,
            FTag.LegCFICode,
            FTag.LegSecurityType,
            FTag.LegSecuritySubType,
            FTag.LegMaturityMonthYear,
            FTag.LegMaturityDate,
            FTag.LegCouponPaymentDate,
            FTag.LegIssueDate,
            FTag.LegRepoCollateralSecurityType,
            FTag.LegRepurchaseTerm,
            FTag.LegRepurchaseRate,
            FTag.LegFactor,
            FTag.LegCreditRating,
            FTag.LegInstrRegistry,
            FTag.LegCountryOfIssue,
            FTag.LegStateOrProvinceOfIssue,
            FTag.LegLocaleOfIssue,
            FTag.LegRedemptionDate,
            FTag.LegStrikePrice,
            FTag.LegStrikeCurrency,
            FTag.LegOptAttribute,
            FTag.LegContractMultiplier,
            FTag.LegCouponRate,
            FTag.LegSecurityExchange,
            FTag.LegIssuer,
            FTag.EncodedIssuerLen,
            FTag.EncodedLegIssuer,
            FTag.LegSecurityDesc,
            FTag.EncodedLegSecurityDescLen,
            FTag.EncodedLegSecurityDesc,
            FTag.LegRatioQty,
            FTag.LegSide,
            FTag.LegCurrency,
            FTag.LegPool,
            FTag.LegDatedDate,
            FTag.LegContractSettlMonth,
            FTag.LegInterestAccrualDate,
        ],
    }

    def logon(self) -> FIXMessage:
        msg = FIXMessage(FMsg.LOGON)
        msg.set(FTag.EncryptMethod, 0)
        msg.set(FTag.HeartBtInt, 30)
        return msg

    def logout(self) -> FIXMessage:
        msg = FIXMessage(FMsg.LOGOUT)
        return msg

    def heartbeat(self) -> FIXMessage:
        msg = FIXMessage(FMsg.HEARTBEAT)
        return msg

    def test_request(self) -> FIXMessage:
        msg = FIXMessage(FMsg.TESTREQUEST)
        return msg

    def sequence_reset(self, responding_to, is_gap_fill) -> FIXMessage:
        msg = FIXMessage(FMsg.SEQUENCERESET)
        msg.set(FTag.GapFillFlag, "Y" if is_gap_fill else "N")
        msg.set(FTag.MsgSeqNum, responding_to[FTag.BeginSeqNo])
        return msg

    def resend_request(self, begin_seq_no, end_seq_no="0") -> FIXMessage:
        msg = FIXMessage(FMsg.RESENDREQUEST)
        msg.set(FTag.BeginSeqNo, str(begin_seq_no))
        msg.set(FTag.EndSeqNo, str(end_seq_no))
        return msg



import pytest

from asyncfix.protocol.common import FExecType, FOrdSide, FOrdStatus, FOrdType


def test_exec_type():
    assert FExecType.NEW == "0"
    assert FExecType.NEW in {"0"}
    assert "0" in {FExecType.NEW}


def test_ord_size():
    assert FOrdSide.BUY == "1"
    assert FOrdSide.SELL == "2"
    assert FOrdSide.BUY in {"1"}
    assert "1" in {FOrdSide.BUY}


def test_ord_status():
    assert FOrdStatus.CREATED == "Z"  # Non FIX! Only for internal use
    assert FOrdStatus.NEW == "0"
    assert FOrdStatus.NEW in {"0"}
    assert "0" in {FOrdStatus.NEW}
